from kivy.app import Builder

from ubo_gui.page import PAGE_MAX_ITEMS

PAGE_SIZE = PAGE_MAX_ITEMS
SHORT_WIDTH = 46

Builder.load_string(
    f"""
#:set UBO_GUI_PAGE_SIZE '{PAGE_SIZE}'
#:set UBO_GUI_SHORT_WIDTH '{SHORT_WIDTH}'
""",
)



"""Module for the `HeaderMenuPageWidget` class."""
from __future__ import annotations

import pathlib
import warnings
from typing import TYPE_CHECKING, Any, Sequence

from kivy.app import Builder, StringProperty

from ubo_gui.menu.constants import PAGE_SIZE
from ubo_gui.page import PageWidget

if TYPE_CHECKING:
    from ubo_gui.menu.types import Item


class HeaderMenuPageWidget(PageWidget):
    """Renders a header page of a `Menu`."""

    heading = StringProperty()
    sub_heading = StringProperty()

    def __init__(
        self: HeaderMenuPageWidget,
        items: Sequence[Item],
        heading: str = '',
        sub_heading: str = '',
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        """Initialize a `HeaderMenuPageWidget`.

        Parameters
        ----------
        items: `Sequence`[[`Item`]]
            The item to be shown in this page

        heading: `str`
            The heading of the page

        sub_heading: `str`
            The sub-heading of the page

        kwargs: Any
            Stuff that will get directly passed to the `__init__` method of Kivy's
        `Screen`.

        """
        if len(items) > 1:
            msg = '`HeaderMenuPageWidget` is initialized with more than one item'
            raise ValueError(msg)
        super().__init__(items, **kwargs)
        self.heading = heading
        self.sub_heading = sub_heading

    def get_item(self: HeaderMenuPageWidget, index: int) -> Item | None:
        """Get the item at the given index."""
        if index != PAGE_SIZE - 1:
            warnings.warn(
                f'index must be {PAGE_SIZE - 1}',
                ResourceWarning,
                stacklevel=1,
            )
            return None
        try:
            return self.items[index - PAGE_SIZE + 1]
        except IndexError:
            return None


Builder.load_file(
    pathlib.Path(__file__)
    .parent.joinpath('header_menu_page_widget.kv')
    .resolve()
    .as_posix(),
)



"""Module containing the `PageWidget` class."""
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Self, Sequence

from kivy.properties import ListProperty, StringProperty
from kivy.uix.screenmanager import Screen

if TYPE_CHECKING:
    from menu.types import Item


PAGE_MAX_ITEMS = 3


class PageWidget(Screen):
    """A `Screen` that represents a page in the menu."""

    __events__ = ('on_close',)

    items: Sequence[Item] = ListProperty([])
    title: str | None = StringProperty(allownone=True, defaultvalue=None)

    def go_up(self: Self) -> None:
        """Implement this method to provide custom logic for up key."""
        return

    def go_down(self: Self) -> None:
        """Implement this method to provide custom logic for down key."""
        return

    def go_back(self: Self) -> bool | None:
        """Implement this method to provide custom logic for back key.

        Return `true` if back button is handled and the default behavior should be
        avoided
        """
        return

    def __init__(
        self: PageWidget,
        items: Sequence[Item] | None = None,
        *args: object,
        **kwargs: object,
    ) -> None:
        """Initialize a `PageWidget`.

        Parameters
        ----------
        items: `list` of `Item`
            The items to be shown in this page

        args: object
            Stuff that will get directly passed to the `__init__` method of Kivy's

        kwargs: object
            Stuff that will get directly passed to the `__init__` method of Kivy's
        `Screen`.
        """
        if items and len(items) > PAGE_MAX_ITEMS:
            msg = f"""`PageWidget` is initialized with more than `MAX_ITEMS`={
            PAGE_MAX_ITEMS} items"""
            raise ValueError(msg)
        self.items = items or []
        super().__init__(*args, **kwargs)

    def get_item(self: PageWidget, index: int) -> Item | None:
        """Get the page item at the given index."""
        if not 0 <= index < len(self.items):
            msg = f"""index must be greater than or equal to 0 and less than {
            len(self.items)}"""
            warnings.warn(msg, ResourceWarning, stacklevel=1)
            return None
        return self.items[index]

    def on_close(self: PageWidget) -> None:
        """Signal when the page is closed."""

    def on_leave(self: PageWidget, *args: object) -> None:
        """Override `on_leave` to dispatch `on_close` event."""
        self.dispatch('on_close')
        return super().on_leave(*args)



"""
This module initializes the ubo_gui package and registers various widgets with the Kivy Factory.
It also defines the paths for assets and fonts used by the package.
"""
from pathlib import Path

from kivy.factory import Factory

__import__('ubo_gui.constants')
__import__('ubo_gui.menu.constants')

ROOT_PATH = Path(__file__).parent
ASSETS_PATH = ROOT_PATH.joinpath('assets')
FONTS_PATH = ASSETS_PATH.joinpath('fonts')


Factory.register('AnimatedSlider', module='ubo_gui.animated_slider')
Factory.register('GaugeWidget', module='ubo_gui.gauge')
Factory.register('ItemWidget', module='ubo_gui.menu.widgets.item_widget')
Factory.register('MenuWidget', module='ubo_gui.menu')
Factory.register('NotificationWidget', module='ubo_gui.notification')
Factory.register('PromptWidget', module='ubo_gui.prompt')
Factory.register('VolumeWidget', module='ubo_gui.volume')



from __future__ import annotations

import pathlib
import warnings
from typing import (
    TYPE_CHECKING,
    Callable,
)

from kivy.app import Builder
from kivy.metrics import dp
from kivy.properties import ColorProperty, ObjectProperty, StringProperty

from ubo_gui.constants import DANGER_COLOR
from ubo_gui.menu.types import ActionItem
from ubo_gui.page import PAGE_MAX_ITEMS, PageWidget

if TYPE_CHECKING:
    from ubo_gui.menu.types import Item


class NotificationAction:
    title: str
    callback: Callable

    def __init__(self: NotificationAction, *, title: str, callback: Callable) -> None:
        self.title = title
        self.callback = callback

    def __repr__(self: NotificationAction) -> str:
        """Return a string representation of this object."""
        return f'<NotificationAction(title={self.title}, callback={self.callback})>'


class NotificationWidget(PageWidget):
    """renders a notification."""

    __events__ = ('on_dismiss',)

    notification = ObjectProperty()
    notification_title: str = StringProperty()
    content: str = StringProperty()
    icon: str = StringProperty()
    color = ColorProperty()

    def __init__(
        self: NotificationWidget,
        *args: object,
        **kwargs: object,
    ) -> None:
        super().__init__(
            *args,
            items=[
                ActionItem(
                    icon='delete',
                    action=lambda: self.dispatch('on_dismiss') and None,
                    label='',
                    is_short=True,
                    background_color=DANGER_COLOR,
                ),
            ],
            **kwargs,
        )

    def go_down(self: NotificationWidget) -> None:
        """Scroll down the notification list."""
        self.ids.slider.animated_value -= dp(50)

    def go_up(self: NotificationWidget) -> None:
        """Scroll up the notification list."""
        self.ids.slider.animated_value += dp(50)

    def get_item(self: NotificationWidget, index: int) -> Item | None:
        """Get the page item at the given index."""
        if index != PAGE_MAX_ITEMS - 1:
            warnings.warn('index must be 2', ResourceWarning, stacklevel=1)
            return None
        return self.items[index - 2]

    def on_dismiss(self: PageWidget) -> None:
        """Signal when the notification is dismissed."""
        ...


Builder.load_file(
    pathlib.Path(__file__)
    .parent.joinpath('notification_widget.kv')
    .resolve()
    .as_posix(),
)



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import logging
import sys

logger = logging.getLogger('ubo-gui')
logger.propagate = False


class ExtraFormatter(logging.Formatter):
    def_keys = (
        'name',
        'msg',
        'args',
        'levelname',
        'levelno',
        'pathname',
        'filename',
        'module',
        'exc_info',
        'exc_text',
        'stack_info',
        'lineno',
        'funcName',
        'created',
        'msecs',
        'relativeCreated',
        'thread',
        'threadName',
        'processName',
        'process',
        'message',
    )

    def format(self: ExtraFormatter, record: logging.LogRecord) -> str:
        string = super().format(record)
        extra = {k: v for k, v in record.__dict__.items() if k not in self.def_keys}
        if len(extra) > 0:
            string += ' - extra: ' + str(extra)

        return string


def add_stdout_handler(level: int = logging.DEBUG) -> None:
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(level)
    stdout_handler.setFormatter(
        ExtraFormatter(
            '%(created)f [%(levelname)s] %(message)s',
            '%Y-%m-%d %H:%M:%S',
        ),
    )
    logger.addHandler(stdout_handler)


def add_file_handler(level: int = logging.DEBUG) -> None:
    file_handler = logging.FileHandler('ubo-gui.log')
    file_handler.setLevel(level)
    file_handler.setFormatter(
        ExtraFormatter(
            '%(created)f [%(levelname)s] %(message)s',
            '%Y-%m-%d %H:%M:%S',
        ),
    )
    logger.addHandler(file_handler)


__all__ = ('logger', 'add_stdout_handler', 'add_file_handler')



"""Provides easy access to different transitions."""
from __future__ import annotations

from functools import cached_property

from headless_kivy_pi import HeadlessWidget
from kivy.uix.screenmanager import (
    NoTransition,
    SlideTransition,
    SwapTransition,
    TransitionBase,
)


class TransitionsMixin:
    """Provides easy access to different transitions."""

    def _handle_transition_progress(
        self: TransitionsMixin,
        transition: TransitionBase,
        progression: float,
    ) -> None:
        if progression is 0:  # noqa: F632 - float 0.0 is not accepted, we are looking for int 0
            HeadlessWidget.activate_high_fps_mode()
        transition.screen_out.opacity = 1 - progression
        transition.screen_in.opacity = progression

    def _handle_transition_complete(
        self: TransitionsMixin,
        transition: TransitionBase,
    ) -> None:
        transition.screen_out.opacity = 0
        transition.screen_in.opacity = 1
        HeadlessWidget.activate_low_fps_mode()

    def _setup_transition(self: TransitionsMixin, transition: TransitionBase) -> None:
        transition.bind(on_progress=self._handle_transition_progress)
        transition.bind(on_complete=self._handle_transition_complete)

    @cached_property
    def _no_transition(self: TransitionsMixin) -> NoTransition:
        transition = NoTransition()
        self._setup_transition(transition)
        return transition

    @cached_property
    def _slide_transition(self: TransitionsMixin) -> SlideTransition:
        transition = SlideTransition()
        self._setup_transition(transition)
        return transition

    @cached_property
    def _swap_transition(self: TransitionsMixin) -> SwapTransition:
        transition = SwapTransition()
        self._setup_transition(transition)
        return transition



"""Module for the `AnimatedSlider` class."""

from __future__ import annotations

from kivy.animation import Animation
from kivy.properties import BoundedNumericProperty
from kivy.uix.slider import Slider


class AnimatedSlider(Slider):
    """A slider that moves up and down when its value is changed."""

    animated_value = BoundedNumericProperty(0)

    def animated_value_error_handler(self: AnimatedSlider, value: float) -> float:
        """Handle error for the `animated_value` property's `errorhandler` argument.

        This method is called when the `animated_value` property is set to a
        value outside of its `min` and `max` bounds.
        This method returns the value that the `animated_value` property should be
        set to, instead of the value that was passed to it.
        """
        if value < self.min:
            return self.min
        if value > self.max:
            return self.max
        return 0

    def __init__(self: AnimatedSlider, **kwargs: object) -> None:
        """Initialize the `AnimatedSlider` class."""
        super().__init__(**kwargs)
        self.value = self.animated_value
        self.padding = 0
        self.property('animated_value').__init__(
            0,
            min=self.min,
            max=self.max,
            errorhandler=self.animated_value_error_handler,
        )

    def on_min(self: AnimatedSlider, *largs: float) -> None:
        """Handle the `min` property being set to a new value."""
        value = largs[1]
        self.property('animated_value').set_min(self, value)

    def on_max(self: AnimatedSlider, *largs: float) -> None:
        """Handle the `max` property being set to a new value."""
        value = largs[1]
        self.property('animated_value').set_max(self, value)

    def on_animated_value(
        self: AnimatedSlider,
        _: AnimatedSlider,
        new_value: float,
    ) -> None:
        """Handle the `animated_value` property being set to a new value.

        Animates the slider moving to the new value.

        Arguments:
        ---------
            new_value: The new value that the `animated_value` property is being set to.

        """
        Animation(value=new_value, duration=0.2).start(self)



from kivy.app import Builder


PRIMARY_COLOR = '#68B7FF'
SECONDARY_COLOR = '#363F4B'
TEXT_COLOR = '#FFFFFF'

DANGER_COLOR = '#FF3F51'
SUCCESS_COLOR = '#03F7AE'

Builder.load_string(
    f"""
#:set UBO_GUI_PRIMARY_COLOR '{PRIMARY_COLOR}'
#:set UBO_GUI_SECONDARY_COLOR '{SECONDARY_COLOR}'
#:set UBO_GUI_TEXT_COLOR '{TEXT_COLOR}'
#:set UBO_GUI_DANGER_COLOR '{DANGER_COLOR}'
#:set UBO_GUI_SUCCESS_COLOR '{SUCCESS_COLOR}'
""",
)



"""Module for the `NormalMenuPageWidget` class."""
import pathlib

from kivy.app import Builder

from ubo_gui.page import PageWidget


class NormalMenuPageWidget(PageWidget):
    """renders a normal page of a `Menu`."""


Builder.load_file(
    pathlib.Path(__file__)
    .parent.joinpath('normal_menu_page_widget.kv')
    .resolve()
    .as_posix(),
)



"""Implement `ItemWidget`."""
from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING, Any, Callable

from kivy.app import Builder
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import (
    BooleanProperty,
    ColorProperty,
    ObjectProperty,
    StringProperty,
)

from ubo_gui.constants import PRIMARY_COLOR
from ubo_gui.menu.types import process_subscribable_value

if TYPE_CHECKING:
    from kivy.graphics import Color

    from ubo_gui.menu.types import Item


class ItemWidget(BoxLayout):
    """Renders an `Item`.

    Attributes
    ----------
    color: `tuple` of `int`
        Color of the item, used for its text and icon.

    background_color: `tuple` of `int`
        Background color of the item.

    label:
        The text rendered in the item.

    icon: `str`
        Name of a Material Symbols icon.

    """

    _subscriptions: list[Callable[[], None]]

    is_set: bool = BooleanProperty(defaultvalue=False)
    label: str = StringProperty()
    color: Color = ColorProperty((1, 1, 1, 1))
    background_color: Color = ColorProperty(PRIMARY_COLOR)
    icon: str = StringProperty(defaultvalue='')
    is_short: bool = BooleanProperty(defaultvalue=False)
    item: Item = ObjectProperty()

    def __init__(self: ItemWidget, **kwargs: dict[str, Any]) -> None:
        """Initialize an `ItemWidget`."""
        super().__init__(**kwargs)
        self._subscriptions = []

    def __del__(self: ItemWidget) -> None:
        """Unsubscribe from the item."""
        self.clear_subscriptions()

    def on_item(self: ItemWidget, instance: ItemWidget, value: Item | None) -> None:
        """Update the widget properties when the item changes."""
        self.clear_subscriptions()
        if value is not None:
            instance.is_set = True
            instance.label = ''
            self._subscriptions.append(
                process_subscribable_value(
                    value.label,
                    lambda value: setattr(instance, 'label', value or ''),
                ),
            )

            instance.is_short = False
            self._subscriptions.append(
                process_subscribable_value(
                    value.is_short,
                    lambda value: setattr(
                        instance,
                        'is_short',
                        False if value is None else value,
                    ),
                ),
            )

            instance.color = ItemWidget.color.defaultvalue
            self._subscriptions.append(
                process_subscribable_value(
                    value.color,
                    lambda value: setattr(
                        instance,
                        'color',
                        value or ItemWidget.color.defaultvalue,
                    ),
                ),
            )

            instance.background_color = ItemWidget.background_color.defaultvalue
            self._subscriptions.append(
                process_subscribable_value(
                    value.background_color,
                    lambda value: setattr(
                        instance,
                        'background_color',
                        value or ItemWidget.background_color.defaultvalue,
                    ),
                ),
            )

            instance.icon = ''
            self._subscriptions.append(
                process_subscribable_value(
                    value.icon,
                    lambda value: setattr(instance, 'icon', value or ''),
                ),
            )
        else:
            instance.is_set = False

    def clear_subscriptions(self: ItemWidget) -> None:
        """Clear the subscriptions."""
        for subscription in self._subscriptions:
            subscription()
        self._subscriptions.clear()


Builder.load_file(
    pathlib.Path(__file__).parent.joinpath('item_widget.kv').resolve().as_posix(),
)
